@@ -2,7 +2,7 @@ cat >> $HGRCPATH <<EOF
 [extensions]
 narrow=
 [ui]
-ssh=python "$RUNTESTDIR/dummyssh"
+ssh=$PYTHON "$RUNTESTDIR/dummyssh"
 [experimental]
 changegroup3 = True
 EOF
